@@ -50,6 +50,13 @@ GHashTable* netdefs;
  * existing definition */
 GHashTable* ids_in_file;
 
+/* List of "seen" ids not found in netdefs yet by the parser.
+ * These are removed when it exists in this list and we reach the point of
+ * creating a netdef for that id; so by the time we're done parsing the yaml
+ * document it should be empty. */
+GHashTable *missing_id;
+int missing_ids_found;
+
 /****************************************************
  * Loading and error handling
  ****************************************************/
@@ -137,6 +144,22 @@ scalar(const yaml_node_t* node)
     return (const char*) node->data.scalar.value;
 }
 
+static void
+add_missing_node(const yaml_node_t* node)
+{
+    missing_node* missing;
+
+    /* Let's capture the current netdef we were playing with along with the
+     * actual yaml_node_t that errors (that is an identifier not previously
+     * seen by the compiler). We can use it later to write an sensible error
+     * message and point the user in the right direction. */
+    missing = g_new0(missing_node, 1);
+    missing->netdef_id = cur_netdef->id;
+    missing->node = node;
+
+    g_debug("recording missing yaml_node_t %s", scalar(node));
+    g_hash_table_insert(missing_id, (gpointer)scalar(node), missing);
+}
 
 /**
  * Check that node contains a valid ID/interface name. Raise GError if not.
@@ -276,11 +299,11 @@ handle_netdef_id_ref(yaml_document_t* doc, yaml_node_t* node, const void* data,
     net_definition* ref = NULL;
 
     ref = g_hash_table_lookup(netdefs, scalar(node));
-    if (!ref)
-        return yaml_error(node, error, "%s: interface %s is not defined",
-                          cur_netdef->id, scalar(node));
-
-    *((net_definition**) ((void*) cur_netdef + offset)) = ref;
+    if (!ref) {
+        add_missing_node(node);
+    } else {
+        *((net_definition**) ((void*) cur_netdef + offset)) = ref;
+    }
     return TRUE;
 }
 
@@ -563,14 +586,15 @@ handle_interfaces(yaml_document_t* doc, yaml_node_t* node, const void* data, GEr
 
         assert_type(entry, YAML_SCALAR_NODE);
         component = g_hash_table_lookup(netdefs, scalar(entry));
-        if (!component)
-            return yaml_error(node, error, "%s: interface %s is not defined",
-                              cur_netdef->id, scalar(entry));
-        component_ref_ptr = ((char**) ((void*) component + GPOINTER_TO_UINT(data)));
-        if (*component_ref_ptr)
-            return yaml_error(node, error, "%s: interface %s is already assigned to %s",
-                              cur_netdef->id, scalar(entry), *component_ref_ptr);
-        *component_ref_ptr = cur_netdef->id;
+        if (!component) {
+            add_missing_node(entry);
+        } else {
+            component_ref_ptr = ((char**) ((void*) component + GPOINTER_TO_UINT(data)));
+            if (*component_ref_ptr && *component_ref_ptr != cur_netdef->id)
+                return yaml_error(node, error, "%s: interface %s is already assigned to %s",
+                                  cur_netdef->id, scalar(entry), *component_ref_ptr);
+            *component_ref_ptr = cur_netdef->id;
+        }
     }
 
     return TRUE;
@@ -700,6 +724,65 @@ handle_routes_metric(yaml_document_t* doc, yaml_node_t* node, const void* _, GEr
     return TRUE;
 }
 
+/****************************************************
+ * Grammar and handlers for network config "bridge_params" entry
+ ****************************************************/
+
+static gboolean
+handle_bridge_path_cost(yaml_document_t* doc, yaml_node_t* node, const void* data, GError** error)
+{
+    for (yaml_node_pair_t* entry = node->data.mapping.pairs.start; entry < node->data.mapping.pairs.top; entry++) {
+        yaml_node_t* key, *value;
+        guint v;
+        gchar* endptr;
+        net_definition *component;
+        guint* ref_ptr;
+
+        key = yaml_document_get_node(doc, entry->key);
+        assert_type(key, YAML_SCALAR_NODE);
+        value = yaml_document_get_node(doc, entry->value);
+        assert_type(value, YAML_SCALAR_NODE);
+
+        component = g_hash_table_lookup(netdefs, scalar(key));
+        if (!component) {
+            add_missing_node(key);
+        } else {
+            ref_ptr = ((guint*) ((void*) component + GPOINTER_TO_UINT(data)));
+            if (*ref_ptr)
+                return yaml_error(node, error, "%s: interface %s already has a path cost of %u",
+                                  cur_netdef->id, scalar(key), *ref_ptr);
+
+            v = g_ascii_strtoull(scalar(value), &endptr, 10);
+            if (*endptr != '\0' || v > G_MAXUINT)
+                return yaml_error(node, error, "invalid unsigned int value %s", scalar(value));
+
+            g_debug("%s: adding path '%s' of cost: %d", cur_netdef->id, scalar(key), v);
+
+            *ref_ptr = v;
+        }
+    }
+    return TRUE;
+}
+
+const mapping_entry_handler bridge_params_handlers[] = {
+    {"ageing-time", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bridge_params.ageing_time)},
+    {"priority", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bridge_params.priority)},
+    {"forward-delay", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bridge_params.forward_delay)},
+    {"hello-time", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bridge_params.hello_time)},
+    {"max-age", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bridge_params.max_age)},
+    {"path-cost", YAML_MAPPING_NODE, handle_bridge_path_cost, NULL, netdef_offset(bridge_params.path_cost)},
+    {"stp", YAML_SCALAR_NODE, handle_netdef_bool, NULL, netdef_offset(bridge_params.stp)},
+    {NULL}
+};
+
+static gboolean
+handle_bridge(yaml_document_t* doc, yaml_node_t* node, const void* _, GError** error)
+{
+    cur_netdef->custom_bridging = TRUE;
+    cur_netdef->bridge_params.stp = TRUE;
+    return process_mapping(doc, node, bridge_params_handlers, error);
+}
+
 /****************************************************
  * Grammar and handlers for network config "routes" entry
  ****************************************************/
@@ -740,6 +823,70 @@ handle_routes(yaml_document_t* doc, yaml_node_t* node, const void* _, GError** e
     return TRUE;
 }
 
+/****************************************************
+ * Grammar and handlers for bond parameters
+ ****************************************************/
+
+static gboolean
+handle_arp_ip_targets(yaml_document_t* doc, yaml_node_t* node, const void* _, GError** error)
+{
+    for (yaml_node_item_t *i = node->data.sequence.items.start; i < node->data.sequence.items.top; i++) {
+        struct in_addr a4;
+        int ret;
+        g_autofree char* addr = NULL;
+        yaml_node_t *entry = yaml_document_get_node(doc, *i);
+        assert_type(entry, YAML_SCALAR_NODE);
+
+        addr = g_strdup(scalar(entry));
+
+        /* is it an IPv4 address? */
+        ret = inet_pton(AF_INET, addr, &a4);
+        g_assert(ret >= 0);
+        if (ret > 0) {
+            if (!cur_netdef->bond_params.arp_ip_targets)
+                cur_netdef->bond_params.arp_ip_targets = g_array_new(FALSE, FALSE, sizeof(char*));
+            char* s = g_strdup(scalar(entry));
+            g_array_append_val(cur_netdef->bond_params.arp_ip_targets, s);
+            continue;
+        }
+
+        return yaml_error(node, error, "malformed address '%s', must be X.X.X.X or X:X:X:X:X:X:X:X", scalar(entry));
+    }
+
+    return TRUE;
+}
+
+const mapping_entry_handler bond_params_handlers[] = {
+    {"mode", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.mode)},
+    {"lacp-rate", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.lacp_rate)},
+    {"mii-monitor-interval", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.monitor_interval)},
+    {"min-links", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.min_links)},
+    {"transmit-hash-policy", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.transmit_hash_policy)},
+    {"ad-select", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.selection_logic)},
+    {"all-slaves-active", YAML_SCALAR_NODE, handle_netdef_bool, NULL, netdef_offset(bond_params.all_slaves_active)},
+    {"arp-interval", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.arp_interval)},
+    /* TODO: arp_ip_targets */
+    {"arp-ip-targets", YAML_SEQUENCE_NODE, handle_arp_ip_targets},
+    {"arp-validate", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.arp_validate)},
+    {"arp-all-targets", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.arp_all_targets)},
+    {"up-delay", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.up_delay)},
+    {"down-delay", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.down_delay)},
+    {"fail-over-mac-policy", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.fail_over_mac_policy)},
+    {"gratuitious-arp", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.gratuitious_arp)},
+    /* TODO: unsolicited_na */
+    {"packets-per-slave", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.packets_per_slave)},
+    {"primary-reselect-policy", YAML_SCALAR_NODE, handle_netdef_str, NULL, netdef_offset(bond_params.primary_reselect_policy)},
+    {"resend-igmp", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.resend_igmp)},
+    {"learn-packet-interval", YAML_SCALAR_NODE, handle_netdef_guint, NULL, netdef_offset(bond_params.learn_interval)},
+    {NULL}
+};
+
+static gboolean
+handle_bonding(yaml_document_t* doc, yaml_node_t* node, const void* _, GError** error)
+{
+    return process_mapping(doc, node, bond_params_handlers, error);
+}
+
 /****************************************************
  * Grammar and handlers for network devices
  ****************************************************/
@@ -790,6 +937,7 @@ const mapping_entry_handler bridge_def_handlers[] = {
     {"gateway6", YAML_SCALAR_NODE, handle_gateway6},
     {"nameservers", YAML_MAPPING_NODE, NULL, nameservers_handlers},
     {"interfaces", YAML_SEQUENCE_NODE, handle_interfaces, NULL, netdef_offset(bridge)},
+    {"parameters", YAML_MAPPING_NODE, handle_bridge},
     {"routes", YAML_SEQUENCE_NODE, handle_routes},
     {NULL}
 };
@@ -803,7 +951,9 @@ const mapping_entry_handler bond_def_handlers[] = {
     {"gateway6", YAML_SCALAR_NODE, handle_gateway6},
     {"nameservers", YAML_MAPPING_NODE, NULL, nameservers_handlers},
     {"interfaces", YAML_SEQUENCE_NODE, handle_interfaces, NULL, netdef_offset(bond)},
+    {"macaddress", YAML_SCALAR_NODE, handle_netdef_mac, NULL, netdef_offset(set_mac)},
     {"routes", YAML_SEQUENCE_NODE, handle_routes},
+    {"parameters", YAML_MAPPING_NODE, handle_bonding},
     {NULL}
 };
 
@@ -842,8 +992,15 @@ handle_network_renderer(yaml_document_t* doc, yaml_node_t* node, const void* _,
 static gboolean
 validate_netdef(net_definition* nd, yaml_node_t* node, GError** error)
 {
+    int missing_id_count = g_hash_table_size(missing_id);
     g_assert(nd->type != ND_NONE);
 
+    /* Skip all validation if we're missing some definition IDs (devices).
+     * The ones we have yet to see may be necessary for validation to succeed,
+     * we can complete it on the next parser pass. */
+    if (missing_id_count > 0)
+        return TRUE;
+
     /* set-name: requires match: */
     if (nd->set_name && !nd->has_match)
         return yaml_error(node, error, "%s: set-name: requires match: properties", nd->id);
@@ -893,6 +1050,12 @@ handle_network_type(yaml_document_t* doc, yaml_node_t* node, const void* data, G
 
         assert_type(value, YAML_MAPPING_NODE);
 
+        /* At this point we've seen a new starting definition, if it has been
+         * already mentioned in another netdef, removing it from our "missing"
+         * list. */
+        if(g_hash_table_remove(missing_id, scalar(key)))
+            missing_ids_found++;
+
         cur_netdef = g_hash_table_lookup(netdefs, scalar(key));
         if (cur_netdef) {
             /* already exists, overriding/amending previous definition */
@@ -909,8 +1072,9 @@ handle_network_type(yaml_document_t* doc, yaml_node_t* node, const void* data, G
             g_hash_table_insert(netdefs, cur_netdef->id, cur_netdef);
         }
 
-        if (!g_hash_table_add(ids_in_file, cur_netdef->id))
-            return yaml_error(key, error, "Duplicate net definition ID '%s'", cur_netdef->id);
+        // XXX: breaks multi-pass parsing.
+        //if (!g_hash_table_add(ids_in_file, cur_netdef->id))
+        //    return yaml_error(key, error, "Duplicate net definition ID '%s'", cur_netdef->id);
 
         /* and fill it with definitions */
         switch (cur_netdef->type) {
@@ -957,7 +1121,55 @@ const mapping_entry_handler root_handlers[] = {
     {NULL}
 };
 
+/**
+ * Handle multiple-pass parsing of the yaml document.
+ */
+static gboolean
+process_document(yaml_document_t* doc, GError** error)
+{
+    gboolean ret;
+    int previously_found;
+    int still_missing;
 
+    g_assert(missing_id == NULL);
+    missing_id = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, g_free);
+
+    do {
+        g_debug("starting new processing pass");
+
+        previously_found = missing_ids_found;
+        missing_ids_found = 0;
+
+        ret = process_mapping(doc, yaml_document_get_root_node(doc), root_handlers, error);
+
+        still_missing = g_hash_table_size(missing_id);
+
+        if (still_missing > 0 && missing_ids_found == previously_found)
+            break;
+    } while (still_missing > 0 || missing_ids_found > 0);
+
+    if (g_hash_table_size(missing_id) > 0) {
+        GHashTableIter iter;
+        gpointer key, value;
+        missing_node *missing;
+
+        g_clear_error(error);
+
+        /* Get the first missing identifier we can get from our list, to
+         * approximate early failure and give the user a meaningful error. */
+        g_hash_table_iter_init (&iter, missing_id);
+        g_hash_table_iter_next (&iter, &key, &value);
+        missing = (missing_node*) value;
+
+        return yaml_error(missing->node, error, "%s: interface %s is not defined",
+                          missing->netdef_id,
+                          key);
+    }
+
+    g_hash_table_destroy(missing_id);
+    missing_id = NULL;
+    return ret;
+}
 
 /**
  * Parse given YAML file and create/update global "netdefs" list.
@@ -981,7 +1193,8 @@ parse_yaml(const char* filename, GError** error)
     g_assert(ids_in_file == NULL);
     ids_in_file = g_hash_table_new(g_str_hash, NULL);
 
-    ret = process_mapping(&doc, yaml_document_get_root_node(&doc), root_handlers, error);
+    ret = process_document(&doc, error);
+
     cur_netdef = NULL;
     yaml_document_delete(&doc);
     g_hash_table_destroy(ids_in_file);
